/*
 * bt_loc_client.c: bluetooth location services client main program
 * Copyright 2003 by Columbia University; all rights reserved
 * Author: Howard Chu (hhc42@cs.columbia.edu)
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <time.h>

#include "bt_loc_lib.h"
#include "bt_loc_obex.h"

int main(int argc, char **argv)
{
  inquiry_info ii[MAX_INQUIRY];
  char dev_bdaddr_str[BDADDR_LEN];
  char des_bdaddr_str[BDADDR_LEN];
  bdaddr_t bdaddr, dev_bdaddr;
  int l2cap_sock;
  int file_size;
  char tmp[5];
  char save_path[500];
  int dev_id;
  int method;
  obex_t *handle;
  uint8_t channel;
  struct context global_context = { 0, };
  time_t start_time, end_time, elapsed;
  char *file_name;
  int sleep_time;

  /* specify a specific device for easier testing purposes */
  if (argc == 4) {
    strcpy(tmp, DEFAULT_DEV_ID);
    dev_id = hci_devid(tmp);
    printf("Device id: %d\n", dev_id);
    method = atoi(argv[2]);

    if (dev_id == -1 || hci_devba(dev_id, &dev_bdaddr) < 0) {
      perror("Error: Device is not available");
      exit(-1);
    }

    ba2str(&dev_bdaddr, dev_bdaddr_str);
    printf("Device address: %s\n", dev_bdaddr_str);
  } else if (argc == 5) {
    strcpy(tmp, argv[4]);
    dev_id = hci_devid(tmp);
    printf("Device id: %d\n", dev_id);
    method = atoi(argv[2]);

    if (dev_id == -1 || hci_devba(dev_id, &dev_bdaddr) < 0) {
      perror("Error: Device is not available");
      exit(-1);
    }

    ba2str(&dev_bdaddr, dev_bdaddr_str);
    printf("Device address: %s\n", dev_bdaddr_str);
  } else {
    printf
	("Usage: %s [sleep time for client (in seconds)] [1 for obex, 2 for sockets] [save directory path] [hci# # is the number of the device (optional)]\n",
	 argv[0]);
    return -1;
  }

  /* make sure the 3rd parameter is a directory path, if not, exit */
  if (argv[3][0] != '/') {
    printf("Invalid directory path\n");
    printf
	("Usage: %s [sleep time for client (in seconds)] [1 for obex, 2 for sockets] [save directory path] [hci# # is the number of the device (optional)]\n",
	 argv[0]);
    return -1;
  }

  /* set the path */
  strcpy(save_path, argv[3]);

  /* set the sleep time of the client */
  sleep_time = atoi(argv[1]);

  while (1) {
    start_time = time(NULL);
    // use socket
    if (method == L2CAP_SOCKET) {
      int tmp;

      // try to get the bluetooth address, if fail, exits abnormally
      if (get_bt_address(dev_id, &bdaddr, ii) < 0) {
	perror("Cannot find server\n");
	exit(-1);
      }

      l2cap_sock = 0;
      if ((l2cap_sock = connect_to_server(bdaddr)) == -1) {
	sleep(sleep_time);
	continue;
      }

      file_size = 0;
      tmp =
	  request_location(l2cap_sock, &file_name, &file_size, save_path);

      // general error handling, error messages are generated
      // within the request_location() method
      if (tmp == -1) {
	return -1;
      }
      // this is the case for which it can't find the location service
      // of this particular server, sleep for a bit and try again.
      else if (tmp == -2) {	// inquire again later
	sleep(sleep_time);
	continue;
      }

      end_time = time(NULL);
      elapsed = end_time - start_time;
      printf("It took %d seconds to retrieve the location information.\n",
	     (int) elapsed);
      start_time = 0, end_time = 0, elapsed = 0;

      /* process the location file by parsing through the given filename */
      if (process_location_file(file_name) == -1) {
	return -1;
      }
      free(file_name);

      close(l2cap_sock);
      sleep(sleep_time);

    }
    /* use obex */
    else if (method == OBEX) {
      int my_channel;
      char filename[30];

      if (get_opush_address(dev_id, &bdaddr, ii, &my_channel) == -1) {
	/* sleep for a bit and try again */
	sleep(sleep_time);
	continue;
      } else {
	int error;
	channel = my_channel;
	ba2str(&bdaddr, des_bdaddr_str);

	if (!
	    (handle =
	     OBEX_Init(OBEX_TRANS_BLUETOOTH, obex_event_handler, 0))) {
	  perror("OBEX_Init failed");
	  exit(-1);
	}

	printf("OBEX_Init succeed.\n");

	OBEX_SetUserData(handle, &global_context);

	error =
	    BtOBEX_TransportConnect(handle, &dev_bdaddr, &bdaddr, channel);
	if (error >= 0) {
	  printf("Connected to obex server on: %s\n", des_bdaddr_str);
	  // get the file
	  client_getfile(handle, &global_context, save_path);
	  disconnect_client(handle);

	  end_time = time(NULL);
	  elapsed = end_time - start_time;
	  printf
	      ("It took %d seconds to retrieve the location information.\n",
	       (int) elapsed);
	  start_time = 0, end_time = 0, elapsed = 0;
	  strcpy(filename, global_context.get_name);

	  if (process_location_file(filename) == -1) {
	    sleep(sleep_time);
	    continue;
	  }
	  // sleep for a bit and get the location again
	  sleep(sleep_time);
	  continue;
	} else {
	  printf("No obex server is available on: %s\n", des_bdaddr_str);
	  sleep(sleep_time);
	  continue;
	}

	// default, shouldn't be down here
	sleep(sleep_time);
	continue;
      }
    }
    // invalid method entries
    else {
      printf("Invalid method entires [1 for obex, 2 for sockets]\n");
      exit(-1);
    }
  }

  return 0;
}
