/*
 * bt_loc_lib.c: bluetooth location services library 
 * Copyright 2003 by Columbia University; all rights reserved
 * Author: Howard Chu (hhc42@cs.columbia.edu)
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <string.h>
#include <netdb.h>
#include <sys/time.h>
#include <unistd.h>

#include "bt_loc_lib.h"
#include "bt_loc_xml.h"

/*
 * connect_to_server()
 * Used by the client to connect to the server at location specified by
 * the bluetooth address. The parameter bdaddr is the address that the
 * client wants to connect to. Returns the socket descriptor if succeed, -1
 * if error. 
 */
int connect_to_server(bdaddr_t bdaddr)
{
  int l2cap_sock;
  struct sockaddr_l2 client_l2addr;

  l2cap_sock = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
  if (l2cap_sock < 0) {
    printf("BT Socket creation failure: %s\n", strerror(errno));
    return -1;
  }

  client_l2addr.l2_family = AF_BLUETOOTH;
  client_l2addr.l2_bdaddr = bdaddr;
  client_l2addr.l2_psm = htobs(SDP_PSM);

  // connecting to socket
  if (connect
      (l2cap_sock, (struct sockaddr *) &client_l2addr,
       sizeof(client_l2addr)) < 0) {
    perror("Cannot connect to socket\n");
    close(l2cap_sock);
    return -1;
  }

  return l2cap_sock;
}

/*
 * read_location_file()
 * Used by the server. This server reads the location file and copy its
 * contents to a buffer specified in the parameter.
 * filename is the name of the file to be opened; file_size is the size
 * of the file to be written; file_content is the pointer to the read
 * buffer. Returns 0 if success, -1 if fail
 */
int read_location_file(char *filename, int *file_size, char **file_content)
{
  FILE *fop;
  int c;

  if ((fop = fopen(filename, "r")) == NULL) {
    printf("Cannot open file: %s\n", filename);
    return -1;
  }

  c = 0;
  while (getc(fop) != EOF)
    c++;

  *file_size = c + 1;
  *file_content = (char *) malloc(sizeof(char) * (c + 1));
  memset(*file_content, '\0', (c + 1));

  // assign each byte of the file into the file_content buffer
  rewind(fop);
  fread(*file_content, sizeof(char), (c + 1), fop);
  fclose(fop);

  return 0;
}

/*
 * request_location()
 * Used by the client to request the location information from the server.
 * sd is the socket descriptor of the connection socket; filename will be
 * used to store the location file name received from the server. file_size
 * is used to store the size of the file. save_path is the directory path
 * that the location file will be saved to. Returns 0 if succeed, -1 if fail
 */
int request_location(int sd, char **filename, int *file_size,
		     char *save_path)
{
  FILE *fop;
  int file_length, full_path_length;
  char *file_content;
  char from_server[50];
  char file_name[MAX_FILE_NAME_LENGTH];
  ssize_t read_val;

  if (sd < 0) {
    printf("Invalid socket descriptor\n");
    return -1;
  }
  // sending file transfer request
  printf("Send the request to server to transfer the file\n");
  if (write(sd, REQUEST_LOCATION, strlen(REQUEST_LOCATION)) == 1) {
    fprintf(stderr, "Cannot send the request to server: %s\n",
	    strerror(errno));
    return -1;
  }
  printf("Request is sent\n");

  // make sure that the server has accepted the request
  if (read(sd, from_server, 50) == 1) {
    fprintf(stderr, "Cannot read file from server: %s\n", strerror(errno));
    return -1;
  }
  from_server[strlen(REQUEST_LOCATION)] = '\0';

  if (strcmp(from_server, REQUEST_LOCATION) != 0) {
    printf("Cannot discover the location service from the server.\n");
    return -2;
  }
  // read the location file name
  if ((read_val = read(sd, file_name, MAX_FILE_NAME_LENGTH)) == 1) {
    fprintf(stderr, "Cannot read file from server: %s\n", strerror(errno));
    return -1;
  }
  file_name[read_val] = '\0';
  printf("File name: %s\n", file_name);

  full_path_length = strlen(save_path);
  full_path_length += strlen(file_name) + 1;

  *filename = (char *) malloc(sizeof(char) * full_path_length);
  strcpy(*filename, save_path);
  strcat(*filename, file_name);

  // read the file length
  file_length = 0;
  printf("Reading the file length from server\n");
  if (read(sd, &file_length, sizeof(file_length)) == 1) {
    fprintf(stderr, "Cannot read file from server: %s\n", strerror(errno));
    return -1;
  }

  printf("File length: %d bytes\n", file_length);

  // start transferring file
  file_content = (char *) malloc(sizeof(char) * (file_length + 1));
  memset(file_content, '\0', (file_length + 1));

  printf("Start transferring file...\nTransferring\n");
  if (read(sd, file_content, (file_length + 1)) == 1) {
    fprintf(stderr, "Cannot start transfering file: %s\n",
	    strerror(errno));
    return -1;
  }

  if ((fop = fopen(*filename, "w")) == NULL) {
    printf("Cannot open file: %s\n", *filename);
    return -1;
  }

  fwrite(file_content, sizeof(char), strlen(file_content), fop);
  printf("File saved: %s\n\n", *filename);

  *file_size = file_length;
  fclose(fop);

  return 0;
}

/*
 * process_location_file
 * Used by the client to process the location file received from the server.
 * It calls xml_process_file() to do the XML parsing.
 * filename is the path of the location file to be processed. Returns the
 * return value of xml_process_file()
 */
int process_location_file(char *filename)
{
  return xml_process_file(filename);
}

/*
 * get_bt_address()
 * Used by both the client and the server. When the client and the server
 * uses l2cap socket connection, client needs this method to get the
 * bluetooth address of the device that runs SDP. The client needs this
 * method to make sure it itself is running SDP.
 * Returns 0 if sdp is found, returns -1 if fail
 */
int get_bt_address(int dev_id, bdaddr_t * bdaddr_ptr, inquiry_info * ii)
{
  char bdaddr_str[20];
  uint8_t inquiry_count;

  // connect to sdp server, if not successful sleep,
  // wakes up every minute
  if (sdp_general_inquiry(ii, MAX_INQUIRY, CONNECT_TIME, &inquiry_count) <
      0) {
    printf("Sdp inquiry failed\n");
    return -1;
  }
  // just get the first address

  *bdaddr_ptr = ii[0].bdaddr;

  ba2str(bdaddr_ptr, bdaddr_str);
  printf("sdp bd address: %s \n", bdaddr_str);

  return 0;
}

/*
 * init_server()
 * Used by the server to initialize itself.
 * sd is used to store the socket descriptor returned. dev_bdaddr
 * is the device bluetooth address of where the server resides.
 * Returns 0 if succeed, return -1 if fail
 */
int init_server(int *sd, bdaddr_t dev_bdaddr)
{
  struct sockaddr_l2 l2addr;

  // create the socket descriptor
  *sd = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
  if (*sd == -1) {
    printf("BT Socket creation failure:%s\n", strerror(errno));
    close(*sd);
    return -1;
  }

  l2addr.l2_family = AF_BLUETOOTH;
  l2addr.l2_bdaddr = dev_bdaddr;
  l2addr.l2_psm = htobs(SDP_PSM);
  if (0 > bind(*sd, (struct sockaddr *) &l2addr, sizeof(l2addr))) {
    SDPERR("BT bind failure:%s", strerror(errno));
    return -1;
  }
  // listen for connection
  listen(*sd, 1);
  printf("Server has started and is listening on psm: %d\n", SDP_PSM);

  return 0;
}

/*
 * process_request()
 * Used by the server program to process the get location file request from
 * the client. This method does a simple validation process with the client
 * which it will only send the file if the client sends a validation string
 * to the server. file_content is the buffer that holds the file's content.
 * file_size is the size of the file content's buffer. server_socket is the
 * socket descriptor of the server. file_name is the name of the location
 * file.
 */
int process_request(char *file_content, int file_size, int server_socket,
		    char *file_name)
{
  // new l2cap connection for the the connected client.
  struct sockaddr_l2 caddr;
  socklen_t len = sizeof(caddr);
  int connect_sock;
  char client_request[strlen(REQUEST_LOCATION) + 1];

  do {
    // accept connection
    connect_sock = accept(server_socket, (struct sockaddr *) &caddr, &len);

    printf("Connected to Client. Validating...\n");

    // send file if accept
    read(connect_sock, client_request, (strlen(REQUEST_LOCATION) + 1));
    client_request[strlen(REQUEST_LOCATION)] = '\0';

    // simple validation process in which it'll check whether the client
    // send the correct request string.
    if (strcmp(client_request, REQUEST_LOCATION) != 0) {
      printf("Validation failed. Closing connection and listen again.\n");
      write(connect_sock, "fail", 4);
      close(connect_sock);
      continue;
    }

    if (write(connect_sock, REQUEST_LOCATION, strlen(REQUEST_LOCATION)) ==
	1) {
      fprintf(stderr, "Cannot send the validation to client: %s\n",
	      strerror(errno));
      close(connect_sock);
      close(server_socket);
      return -1;
    }
    // send the file, first the filename, then size then the content
    printf("Sending file name: %s\n", file_name);
    write(connect_sock, file_name, strlen(file_name) + 1);

    printf("Sending file size: %d bytes\n", file_size);
    write(connect_sock, &file_size, sizeof(file_size));

    printf("Server is sending the file to the client\n");
    write(connect_sock, file_content, strlen(file_content));
    printf("File transfer is completed.\n");

    // close connection and listens again.
    close(connect_sock);
    printf("Server listens for connection again.\n");

  } while (1);
}
