/*
 * bt_loc_lib.h: bluetooth location services library header file
 * Copyright 2003 by Columbia University; all rights reserved
 * Author: Howard Chu (hhc42@cs.columbia.edu)
 */

#ifndef BT_LOC_SERV_H
#define BT_LOC_SERV_H

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/l2cap.h>

#include "sdp_internal.h"

#define REQUEST_LOCATION "bt_location_request"
#define CONNECT_TIME 8
#define MAX_INQUIRY 20
#define SLEEP_TIME 10
#define BDADDR_LEN 20
#define DEFAULT_DEV_ID "hci1"
#define MAX_FILE_NAME_LENGTH 20
#define L2CAP_SOCKET 2
#define OBEX 1

/*
 * connect_to_server()
 * Used by the client to connect to the server at location specified by
 * the bluetooth address. The parameter bdaddr is the address that the
 * client wants to connect to. Returns the socket descriptor if succeed, -1
 * if error. 
 */
int connect_to_server(bdaddr_t bdaddr);

/*
 * request_location()
 * Used by the client to request the location information from the server.
 * sd is the socket descriptor of the connection socket; filename will be
 * used to store the location file name received from the server. file_size
 * is used to store the size of the file. save_path is the directory path
 * that the location file will be saved to. Returns 0 if succeed, -1 if fail
 */
int request_location(int sd, char **filename, int *file_size,
		     char *save_path);

/*
 * process_location_file
 * Used by the client to process the location file received from the server.
 * It calls xml_process() to do the XML parsing.
 * filename is the path of the location file to be processed. Returns the
 * return value of xml_process()
 */
int process_location_file(char *filename);

/*
 * read_location_file()
 * Used by the server. This server reads the location file and copy its
 * contents to a buffer specified in the parameter.
 * filename is the name of the file to be opened; file_size is the size
 * of the file to be written; file_content is the pointer to the read
 * buffer. Returns 0 if success, -1 if fail
 */
int read_location_file(char *filename, int *file_size,
		       char **file_content);

/*
 * get_bt_address()
 * Used by both the client and the server. When the client and the server
 * uses l2cap socket connection, client needs this method to get the
 * bluetooth address of the device that runs SDP. The client needs this
 * method to make sure it itself is running SDP.
 * Returns 0 if sdp is found, returns -1 if fail
 */
int get_bt_address(int dev_id, bdaddr_t * bdaddr_ptr, inquiry_info * ii);

/*
 * init_server()
 * Used by the server to initialize itself.
 * sd is used to store the socket descriptor returned. dev_bdaddr
 * is the device bluetooth address of where the server resides.
 * Returns 0 if succeed, return -1 if fail
 */
int init_server(int *sd, bdaddr_t dev_bdaddr);

/*
 * process_request()
 * Used by the server program to process the get location file request from
 * the client. This method does a simple validation process with the client
 * which it will only send the file if the client sends a validation string
 * to the server. file_content is the buffer that holds the file's content.
 * file_size is the size of the file content's buffer. server_socket is the
 * socket descriptor of the server. file_name is the name of the location
 * file.
 */
int process_request(char *file_content, int file_size, int server_socket,
		    char *file_name);

#endif
