/*
 * bt_loc_obex.c: bluetooth location services obex related functions  
 * Copyright 2003 by Columbia University; all rights reserved
 * Author: Howard Chu (hhc42@cs.columbia.edu)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <libgen.h>
#include <string.h>
#include <sys/stat.h>

#include "bt_loc_obex.h"

int fileDesc;
#define OBEX_STREAM_CHUNK       4096
uint8_t buffer[OBEX_STREAM_CHUNK];

/*
 * do_search()
 * Used by both the server and the client to make sure the sdp server
 * is currently running the OPUSH services so that we can carry out
 * obex object transfer. bdaddr is the pointer to the bluetooth address
 * of the sdp server to be searched; context specifies the service to be
 * searched. opush_channel is used to store the channel of the opush
 * service.
 * Returns 0 if succeed, -1 if fail
 */
int do_search(bdaddr_t * bdaddr, struct search_context *context,
	      int *opush_channel)
{
  sdp_list_t *attrid, *search, *seq, *next;
  uint32_t range = 0x0000ffff;
  char str[20];
  sdp_session_t *sess;

  sess = sdp_connect(BDADDR_ANY, bdaddr, 0);
  ba2str(bdaddr, str);
  if (!sess) {
    printf("Failed to connect to SDP server on %s\n", str);
    return -1;
  }
  if (context->svc)
    printf("Searching for %s on %s ...\n", context->svc, str);
  else
    printf("Browsing %s ...\n", str);

  attrid = sdp_list_append(0, &range);
  search = sdp_list_append(0, &context->group);
  if (sdp_service_search_attr_req
      (sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
    printf("Service Search failed\n");
    sdp_close(sess);
    return -1;
  }
  sdp_list_free(attrid, 0);
  sdp_list_free(search, 0);

  for (; seq; seq = next) {
    sdp_record_t *rec = (sdp_record_t *) seq->data;
    struct search_context sub_context;

    *opush_channel = get_opush_channel(rec);

    if (sdp_get_group_id(rec, &sub_context.group) != -1) {
      /* Set the subcontext for browsing the sub tree */
      memcpy(&sub_context, context, sizeof(struct search_context));
      /* Browse the next level down if not done */
      if (sub_context.group.value.uuid16 != context->group.value.uuid16)
	do_search(bdaddr, &sub_context, opush_channel);
    }
    next = seq->next;
    free(seq);
    sdp_record_free(rec);
  }

  sdp_close(sess);
  return 0;
}

/*
 * get_opush_address()
 * This method gets the bluetooth address of the device that contains
 * the SDP server that's currently running the OPUSH service.
 * dev_id is the device id, bdaddr_ptr is used to store the bluetooth
 * address of the SDP server, ii stores the inquiry information.
 * opush_channel- store the channel of the opush service
 * Returns 0 if succeed, -1 if fail
 */
int get_opush_address(int dev_id, bdaddr_t * bdaddr_ptr, inquiry_info * ii,
		      int *opush_channel)
{
  char bdaddr_str[20];
  uint8_t inquiry_count;
  struct search_context opush;
  uint16_t opush_class;
  opush_class = OBEX_OBJPUSH_SVCLASS_ID;
  opush.svc = strdup("OPUSH");

  sdp_uuid16_create(&opush.group, opush_class);

  // connect to sdp server, if not successful sleep,
  // wakes up every minute
  if (sdp_general_inquiry(ii, MAX_INQUIRY, CONNECT_TIME, &inquiry_count) <
      0) {
    printf("Sdp inquiry failed\n");
    return -1;
  }
  // just get the first address
  *bdaddr_ptr = ii[0].bdaddr;

  if (!do_search(bdaddr_ptr, &opush, opush_channel)) {
    ba2str(bdaddr_ptr, bdaddr_str);
    printf("bd address: %s ", bdaddr_str);
    printf("contains the OPUSH service\n");

    return 0;
  } else {
    ba2str(bdaddr_ptr, bdaddr_str);
    printf("No OPUSH service is found on :%s\n", bdaddr_str);

    return -1;
  }

  return -1;
}

/*
 * client_getfile()
 * A modified version of obex client's getfile function, this allows
 * user to specify a directory path upon saving the file.
 * handle- the obex handle, gt- the obex context, save_path-
 * the directory path that the file is to be saved.
 */
void client_getfile(obex_t * handle, struct context *gt, char *save_path)
{
  obex_object_t *object;
  char rname[200];
  char req_name[200];
  int rname_size;
  obex_headerdata_t hd;
  char path_name[200];

  strcpy(req_name, LOCATION_FILE);

  if ((object = OBEX_ObjectNew(handle, OBEX_CMD_GET)) == NULL) {
    printf("Error\n");
    return;
  }

  rname_size = OBEX_CharToUnicode(rname, req_name, sizeof(rname));

  hd.bs = rname;
  OBEX_ObjectAddHeader(handle, object, OBEX_HDR_NAME, hd,
		       rname_size, OBEX_FL_FIT_ONE_PACKET);

  /* Remember the path of the file we are getting so we can save
     it when we get the response */
  strcpy(path_name, save_path);
  strcat(path_name, req_name);

  gt->get_name = path_name;
  OBEX_Request(handle, object);
  new_syncwait(handle);
  gt->clientdone = TRUE;
}

/*
 * obex_event_handler()
 * Called by the obex-layer when some event occurs. Modified from the origianl
 * obex event handler so that it uses the new version of client get file.
 * handle- obex handle, object- the exchange object, mode- the exchange mode,
 * event- obex event, obex_cmd- obex command, obex_rsp- obex response
 */
void obex_event_handler(obex_t * handle, obex_object_t * object, int mode,
			int event, int obex_cmd, int obex_rsp)
{
  switch (event) {
  case OBEX_EV_PROGRESS:
    printf("Made some progress...\n");
    break;

  case OBEX_EV_ABORT:
    printf("Request aborted!\n");
    break;

  case OBEX_EV_REQDONE:
    if (mode == OBEX_CLIENT) {
      new_client_done(handle, object, obex_cmd, obex_rsp);
    } else {
      server_done(handle, object, obex_cmd, obex_rsp);
    }
    break;
  case OBEX_EV_REQHINT:
    /* Accept any command. Not rellay good, but this is a test-program :) */
    OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
    break;

  case OBEX_EV_REQ:
    server_request(handle, object, event, obex_cmd);
    break;

  case OBEX_EV_LINKERR:
    OBEX_TransportDisconnect(handle);
    printf("Link broken!\n");
    break;

  case OBEX_EV_STREAMEMPTY:
    fillstream(handle, object);
    break;

  default:
    printf("Unknown event %02x!\n", event);
    break;
  }
}

/*
 * new_client_done()
 * A modified version of the original client_done() so that it uses the new
 * saving directory routine.
 * handle- obex handle, object- obex exchange object, obex_cmd- obex command,
 * obex_rsp- obex response
 */
void new_client_done(obex_t * handle, obex_object_t * object, int obex_cmd,
		     int obex_rsp)
{
  struct context *gt;
  gt = OBEX_GetUserData(handle);

  switch (obex_cmd) {
  case OBEX_CMD_CONNECT:
    connect_client_done(handle, object, obex_rsp);
    break;
  case OBEX_CMD_DISCONNECT:
    disconnect_client_done(handle, object, obex_rsp);
    break;
  case OBEX_CMD_PUT:
    put_client_done(handle, object, obex_rsp);
    break;
  case OBEX_CMD_GET:
    new_get_client_done(handle, object, obex_rsp, gt->get_name);
    break;
  case OBEX_CMD_SETPATH:
    setpath_client_done(handle, object, obex_rsp);
    break;
  }
  gt->clientdone = TRUE;
}

/*
 * new_get_client_done()
 * A modified version of the original get_client_done from the obex library
 * so that it supports our get location file routine.
 * handle- obex handle, object- obex exchange object, obex_rsp- obex response,
 * name- the full path of the location file to be saved on the client side
 */
void new_get_client_done(obex_t * handle, obex_object_t * object,
			 int obex_rsp, char *name)
{
  obex_headerdata_t hv;
  uint8_t hi;
  int hlen;

  const uint8_t *body = NULL;
  int body_len = 0;

  if (obex_rsp != OBEX_RSP_SUCCESS) {
    printf("GET failed 0x%02x!\n", obex_rsp);
    return;
  }

  while (OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
    if (hi == OBEX_HDR_BODY) {
      printf("%s() Found body\n", __FUNCTION__);
      body = hv.bs;
      body_len = hlen;
      break;
    } else {
      printf("%s() Skipped header %02x\n", __FUNCTION__, hi);
    }
  }

  if (!body) {
    printf("No body found in answer!\n");
    return;
  }
  printf("GET successful!\n");
  new_safe_save_file(name, body, body_len);
}

/*
 * new_safe_save_file()
 * A modified version of the original safe_save_file() from the obex
 * application library. It will save the exchanged object's buffer
 * to the specified path name.
 * name- full path of the location file to be saved. buf- the buffer
 * to be written to the file. len- the length of the buffer.
 * Returns 0 if succeed, -1 if fail.
 */
int new_safe_save_file(char *name, const uint8_t * buf, int len)
{
  int fd;
  int actual;

  printf("File will be saved at: %s\n", name);

  fd = open(name, O_RDWR | O_CREAT, DEFFILEMODE);

  if (fd < 0) {
    perror(name);
    return -1;
  }

  actual = write(fd, buf, len);
  close(fd);

  printf("Wrote %s (%d bytes)\n", name, actual);

  return actual;
}

/*
 * new_syncwait()
 * This command is used by the obex client's functions to wait for its
 * commands to finish. This method is defined in the obex_test_client.c
 * handle- the obex handle
 */
void new_syncwait(obex_t * handle)
{
  struct context *gt;
  int ret;

  gt = OBEX_GetUserData(handle);

  while (!gt->clientdone) {
    //printf("syncwait()\n");
    ret = OBEX_HandleInput(handle, 10);
    if (ret < 0) {
      printf("Error while doing OBEX_HandleInput()\n");
      break;
    }
    if (ret == 0) {
      /* If running cable. We get no link-errors, so cancel on timeout */
      printf("Timeout waiting for data. Aborting\n");
      OBEX_CancelRequest(handle, FALSE);
      break;
    }
  }

  gt->clientdone = FALSE;
}

/*
 * get_opush_channel()
 * Given a sdp record, we'll look through this record to find the channel
 * with which the opush service is located.
 * rec- sdp record
 * if success, return channel, if fails, return -1
 */
int get_opush_channel(sdp_record_t * rec)
{
  uuid_t *uuid;
  sdp_list_t *list, *list2;

  list = 0;

  if (sdp_get_service_classes(rec, &list) == 0) {

    for (; list; list = list->next) {

      char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
      uuid = (uuid_t *) list->data;

      /* check if this uuid is the opush class's uuid */
      sdp_svclass_uuid2strn(uuid, ServiceClassUUID_str,
			    MAX_LEN_SERVICECLASS_UUID_STR);
      if (strcmp(ServiceClassUUID_str, OPUSH_STR) == 0) {

	if (sdp_get_access_protos(rec, &list2) == 0) {

	  for (; list2; list2 = list2->next) {

	    sdp_list_t *protDescSeq = (sdp_list_t *) list2->data;

	    for (; protDescSeq; protDescSeq = protDescSeq->next) {

	      sdp_data_t *p = (sdp_data_t *) protDescSeq->data;
	      int proto = 0;

	      for (; p; p = p->next) {
		switch (p->dtd) {
		case SDP_UUID16:
		case SDP_UUID32:
		case SDP_UUID128:
		  proto = sdp_uuid_to_proto(&p->val.uuid);
		  break;
		case SDP_UINT8:
		  if (proto == RFCOMM_UUID) {
		    //printf("Channel: %d\n", p->val.uint8);
		    return p->val.uint8;
		  }
		  break;
		}
	      }
	    }

	  }

	  sdp_list_free(list2, (sdp_free_func_t) sdp_data_free);
	}
      }
    }
    sdp_list_free(list, free);
  }

  return -1;
}

/*
 * server_do()
 * This function does the basic server functions and gets called after the 
 * server has been initialized.
 * handle- obex handle
 */
void server_do(obex_t * handle)
{
  struct context *gt;
  gt = OBEX_GetUserData(handle);

  gt->serverdone = FALSE;
  while (!gt->serverdone) {
    if (OBEX_HandleInput(handle, 1) < 0) {
      printf("Error while doing OBEX_HandleInput()\n");
      break;
    }
  }
}

/*
 * connect_client_done()
 * Housekeeping routine after the client is done with its request.
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void connect_client_done(obex_t * handle, obex_object_t * object,
			 int obex_rsp)
{
  uint8_t *nonhdrdata;

  if (obex_rsp == OBEX_RSP_SUCCESS) {
    printf("Connect OK!\n");
    if (OBEX_ObjectGetNonHdrData(object, &nonhdrdata) == 4) {
      printf("Version: 0x%02x. Flags: 0x%02x\n", nonhdrdata[0],
	     nonhdrdata[1]);
    }
  } else {
    printf("Connect failed 0x%02x!\n", obex_rsp);
  }
}

/*
 * disconnect_client()
 * Routine to disconnect client from the transfer
 * handle- obex handle
 */
void disconnect_client(obex_t * handle)
{
  obex_object_t *object;

  if (!(object = OBEX_ObjectNew(handle, OBEX_CMD_DISCONNECT))) {
    printf("Error\n");
    return;
  }

  OBEX_Request(handle, object);
  new_syncwait(handle);
}

/*
 * disconnect_client_done()
 * House keeping routine after client is disconncted.
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void disconnect_client_done(obex_t * handle, obex_object_t * object,
			    int obex_rsp)
{
  printf("Disconnect done!\n");
  OBEX_TransportDisconnect(handle);
}

/*
 * put_client_done()
 * House keeping routine after the put_client operation is done
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void put_client_done(obex_t * handle, obex_object_t * object, int obex_rsp)
{
  if (obex_rsp == OBEX_RSP_SUCCESS) {
    printf("PUT successful!\n");
  } else {
    printf("PUT failed 0x%02x!\n", obex_rsp);
  }
}

/*
 * setpath_client_done()
 * House keeping routine after the setpath_client operation is done
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void setpath_client_done(obex_t * handle, obex_object_t * object,
			 int obex_rsp)
{
  if (obex_rsp == OBEX_RSP_SUCCESS) {
    printf("SETPATH successful!\n");
  } else {
    printf("SETPATH failed 0x%02x!\n", obex_rsp);
  }
}

/*
 * server_request()
 * Routine that the server uses to handle requests from client
 * handle- obex handle, object- obex object, event- event to be handled,
 * cmd- command to handle the request
 */
void server_request(obex_t * handle, obex_object_t * object, int event,
		    int cmd)
{
  switch (cmd) {
  case OBEX_CMD_CONNECT:
    connect_server(handle, object);
    break;
  case OBEX_CMD_DISCONNECT:
    printf("We got a disconnect-request\n");
    OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
    break;
  case OBEX_CMD_GET:
    /* A Get always fits one package */
    get_server(handle, object);
    break;
  case OBEX_CMD_PUT:
    OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
    put_server(handle, object);
    break;
  case OBEX_CMD_SETPATH:
    printf("Got a SETPATH request\n");
    OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
    break;
  default:
    printf("%s() Denied %02x request\n", __FUNCTION__, cmd);
    OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_IMPLEMENTED,
		      OBEX_RSP_NOT_IMPLEMENTED);
    break;
  }
  return;
}

/*
 * server_done()
 * Housekeeping event when the server is done servicing.
 * handle- obex handle, object- obex object, obex_cmd- obex command,
 * obex_rsp- obex response
 */
void server_done(obex_t * handle, obex_object_t * object, int obex_cmd,
		 int obex_rsp)
{
  struct context *gt;
  gt = OBEX_GetUserData(handle);

  printf("Server request finished!\n");

  switch (obex_cmd) {
  case OBEX_CMD_DISCONNECT:
    printf("Disconnect done!\n");
    OBEX_TransportDisconnect(handle);
    gt->serverdone = TRUE;
    break;

  default:
    printf("%s() Command (%02x) has now finished\n", __FUNCTION__,
	   obex_cmd);
    break;
  }
}

/*
 * fillstream()
 * Method used by the client when empty stream needs to be filled in.
 * handle- obex handle, object- obex object
 * Return the bytes filled
 */
int fillstream(obex_t * handle, obex_object_t * object)
{
  int actual;
  obex_headerdata_t hv;

  printf("Filling stream!\n");

  actual = read(fileDesc, buffer, OBEX_STREAM_CHUNK);
  if (actual > 0) {
    /* Read was ok! */
    hv.bs = buffer;
    OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
			 hv, actual, OBEX_FL_STREAM_DATA);
  } else if (actual == 0) {
    /* EOF */
    hv.bs = buffer;
    close(fileDesc);
    fileDesc = -1;
    OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
			 hv, 0, OBEX_FL_STREAM_DATAEND);
  } else {
    /* Error */
    hv.bs = NULL;
    close(fileDesc);
    fileDesc = -1;
    OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY,
			 hv, 0, OBEX_FL_STREAM_DATA);
  }

  return actual;
}

/*
 * connect_server()
 * Routine to connect to the server
 * handle- obex handle, object- obex object
 */
void connect_server(obex_t * handle, obex_object_t * object)
{
  obex_headerdata_t hv;
  uint8_t hi;
  int hlen;

  const uint8_t *who = NULL;
  int who_len = 0;
  printf("%s()\n", __FUNCTION__);

  while (OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
    if (hi == OBEX_HDR_WHO) {
      who = hv.bs;
      who_len = hlen;
    } else {
      printf("%s() Skipped header %02x\n", __FUNCTION__, hi);
    }
  }
  if (who_len == 6) {
    if (strncmp("Linux", who, 6) == 0) {
      printf("Weeeha. I'm talking to the coolest OS ever!\n");
    }
  }
  OBEX_ObjectSetRsp(object, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
}

/*
 * get_server()
 * Routine to get objects from the server
 * handle- obex handle, object- obex object
 */
void get_server(obex_t * handle, obex_object_t * object)
{
  uint8_t *buf;

  obex_headerdata_t hv;
  uint8_t hi;
  int hlen;
  int file_size;

  char *name = NULL;
  char *namebuf = NULL;

  printf("%s()\n", __FUNCTION__);

  while (OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
    switch (hi) {
    case OBEX_HDR_NAME:
      printf("%s() Found name\n", __FUNCTION__);
      if ((namebuf = malloc(hlen / 2))) {
	OBEX_UnicodeToChar(namebuf, hv.bs, hlen);
	name = namebuf;
      }
      break;

    default:
      printf("%s() Skipped header %02x\n", __FUNCTION__, hi);
    }
  }

  if (!name) {
    printf("%s() Got a GET without a name-header!\n", __FUNCTION__);
    OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
    return;
  }
  printf("%s() Got a request for %s\n", __FUNCTION__, name);

  buf = easy_readfile(name, &file_size);
  if (buf == NULL) {
    printf("Can't find file %s\n", name);
    OBEX_ObjectSetRsp(object, OBEX_RSP_NOT_FOUND, OBEX_RSP_NOT_FOUND);
    return;
  }

  OBEX_ObjectSetRsp(object, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
  hv.bs = buf;
  OBEX_ObjectAddHeader(handle, object, OBEX_HDR_BODY, hv, file_size, 0);
  hv.bq4 = file_size;
  OBEX_ObjectAddHeader(handle, object, OBEX_HDR_LENGTH, hv,
		       sizeof(uint32_t), 0);
  free(buf);
  return;
}

/*
 * put_server()
 * This method is responsible for accepting objects
 * handle- obex handle, object- obex object
 */
void put_server(obex_t * handle, obex_object_t * object)
{
  obex_headerdata_t hv;
  uint8_t hi;
  int hlen;

  const uint8_t *body = NULL;
  int body_len = 0;
  char *name = NULL;
  char *namebuf = NULL;

  printf("%s()\n", __FUNCTION__);

  while (OBEX_ObjectGetNextHeader(handle, object, &hi, &hv, &hlen)) {
    switch (hi) {
    case OBEX_HDR_BODY:
      printf("%s() Found body\n", __FUNCTION__);
      body = hv.bs;
      body_len = hlen;
      break;
    case OBEX_HDR_NAME:
      printf("%s() Found name\n", __FUNCTION__);
      if ((namebuf = malloc(hlen / 2))) {
	OBEX_UnicodeToChar(namebuf, hv.bs, hlen);
	name = namebuf;
      }
      break;

    default:
      printf("%s() Skipped header %02x\n", __FUNCTION__, hi);
    }
  }
  if (!body) {
    printf("Got a PUT without a body\n");
    return;
  }
  if (!name) {
    name = "OBEX_PUT_Unknown_object";
    printf("Got a PUT without a name. Setting name to %s\n", name);

  }
  /*
   * ss: Here I can add a function that checks for the name of the file
   * and if the name hasn't been seen before I can send a udp message to the
   * listening daemon
   */
  new_safe_save_file(name, body, body_len);
  free(namebuf);
}

/*
 * easy_readfile()
 * Read a file and alloc a buffer for it
 * filename- name of the file, file_size- size of the file
 * Return the number of bytes read
 */
uint8_t *easy_readfile(const char *filename, int *file_size)
{
  int actual;
  int fd;
  uint8_t *buf;

  *file_size = get_filesize(filename);
  printf("name=%s, size=%d\n", filename, *file_size);

#ifdef _WIN32
  fd = open(filename, O_RDONLY | O_BINARY, 0);
#else
  fd = open(filename, O_RDONLY, 0);
#endif

  if (fd == -1) {
    return NULL;
  }

  if (!(buf = malloc(*file_size))) {
    return NULL;
  }

  actual = read(fd, buf, *file_size);
  close(fd);

#ifdef _WIN32
  if (actual != *file_size) {
    free(buf);
    buf = NULL;
  }
#else
  *file_size = actual;
#endif
  return buf;
}

/*
 * get_filesize()
 * Get the filesize in a "portable" way (Taken from obex-apps)
 * filename- name of the file
 * Return size of the file
 */
int get_filesize(const char *filename)
{
#ifdef _WIN32
  HANDLE fh;
  int size;
  fh = CreateFile(filename, 0, 0, NULL, OPEN_EXISTING, 0, NULL);
  if (fh == INVALID_HANDLE_VALUE) {
    printf("Cannot open %s\n", filename);
    return -1;
  }
  size = GetFileSize(fh, NULL);
  printf("fize size was %d\n", size);
  CloseHandle(fh);
  return size;

#else
  struct stat stats;
  /*  Need to know the file length */
  stat(filename, &stats);
  return (int) stats.st_size;
#endif
}
