/*
 * bt_loc_obex.h: bluetooth location services obex related functions header 
 * Copyright 2003 by Columbia University; all rights reserved
 * Author: Howard Chu (hhc42@cs.columbia.edu)
 */

#ifndef BT_LOC_OBEX_H
#define BT_LOC_OBEX_H

#include <netdb.h>

#include "bt_loc_lib.h"

#include <openobex/obex.h>

#define TRUE 1
#define FALSE 0

#define IR_SERVICE "OBEX"
#define BT_CHANNEL 9
#define LOCATION_FILE "location.xml"
#define OPUSH_STR "OBEX Object Push"

struct context {
  int serverdone;
  int clientdone;
  char *get_name;		/* Name of last get-request */
};

/* Pass args to the inquiry/search handler */
struct search_context {
  char *svc;			/* Service */
  uuid_t group;			/* Browse group */
  int tree;			/* Display full attribute tree */
  uint32_t handle;		/* Service record handle */
};

typedef int (*handler_t) (bdaddr_t * bdaddr, struct search_context * arg);

/*
 * do_search()
 * Used by both the server and the client to make sure the sdp server
 * is currently running the OPUSH services so that we can carry out
 * obex object transfer. bdaddr is the pointer to the bluetooth address
 * of the sdp server to be searched; context specifies the service to be
 * searched. opush_channel- store the channel the opush service is located.
 * Returns 0 if succeed, -1 if fail
 */
int do_search(bdaddr_t * bdaddr, struct search_context *context,
	      int *opush_channel);

/*
 * get_opush_address()
 * This method gets the bluetooth address of the device that contains
 * the SDP server that's currently running the OPUSH service.
 * dev_id is the device id, bdaddr_ptr is used to store the bluetooth
 * address of the SDP server, ii stores the inquiry information.
 * Returns 0 if succeed, -1 if fail
 */
int get_opush_address(int dev_id, bdaddr_t * bdaddr_ptr,
		      inquiry_info * ii, int *opush_channel);

/*
 * client_getfile()
 * A modified version of obex client's getfile function, this allows
 * user to specify a directory path upon saving the file.
 * handle- the obex handle, gt- the obex context, save_path-
 * the directory path that the file is to be saved.
 */
void client_getfile(obex_t * handle, struct context *gt, char *save_path);

/*
 * obex_event_handler()
 * Called by the obex-layer when some event occurs. Modified from the origianl
 * obex event handler so that it uses the new version of client get file.
 * handle- obex handle, object- the exchange object, mode- the exchange mode,
 * event- obex event, obex_cmd- obex command, obex_rsp- obex response
 */
void obex_event_handler(obex_t * handle, obex_object_t * object, int mode,
			int event, int obex_cmd, int obex_rsp);

/*
 * new_client_done()
 * A modified version of the original client_done() so that it uses the new
 * saving directory routine.
 * handle- obex handle, object- obex exchange object, obex_cmd- obex command,
 * obex_rsp- obex response
 */
void new_client_done(obex_t * handle, obex_object_t * object, int obex_cmd,
		     int obex_rsp);

/*
 * new_get_client_done()
 * A modified version of the original get_client_done from the obex library
 * so that it supports our get location file routine.
 * handle- obex handle, object- obex exchange object, obex_rsp- obex response,
 * name- the full path of the location file to be saved on the client side
 */
void new_get_client_done(obex_t * handle, obex_object_t * object,
			 int obex_rsp, char *name);

/*
 * new_safe_save_file()
 * A modified version of the original safe_save_file() from the obex
 * application library. It will save the exchanged object's buffer
 * to the specified path name.
 * name- full path of the location file to be saved. buf- the buffer
 * to be written to the file. len- the length of the buffer.
 * Returns 0 if succeed, -1 if fail.
 */
int new_safe_save_file(char *name, const uint8_t * buf, int len);

/*
 * new_syncwait()
 * This command is used by the obex client's functions to wait for its
 * commands to finish. This method is defined in the obex_test_client.c
 * handle- the obex handle
 */
void new_syncwait(obex_t * handle);

/*
 * get_opush_channel()
 * Given a sdp record, we'll look through this record to find the channel
 * with which the opush service is located.
 * rec- sdp record
 * if success, return channel, if fails, return -1
 */
int get_opush_channel(sdp_record_t * rec);

/*
 * server_do()
 * This function does the basic server functions and gets called after the 
 * server has been initialized.
 * handle- obex handle
 */
void server_do(obex_t * handle);

/*
 * connect_client_done()
 * Housekeeping routine after the client is done with its request.
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void connect_client_done(obex_t * handle, obex_object_t * object,
			 int obex_rsp);

/*
 * disconnect_client()
 * Routine to disconnect client from the transfer
 * handle- obex handle
 */
void disconnect_client(obex_t * handle);

/*
 * disconnect_client_done()
 * House keeping routine after client is disconncted.
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void disconnect_client_done(obex_t * handle, obex_object_t * object,
			    int obex_rsp);

/*
 * put_client_done()
 * House keeping routine after the put_client operation is done
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void put_client_done(obex_t * handle, obex_object_t * object,
		     int obex_rsp);

/*
 * setpath_client_done()
 * House keeping routine after the setpath_client operation is done
 * handle- obex handle, object- obex object, obex_rsp- obex response
 */
void setpath_client_done(obex_t * handle, obex_object_t * object,
			 int obex_rsp);

/*
 * server_request()
 * Routine that the server uses to handle requests from client
 * handle- obex handle, object- obex object, event- event to be handled,
 * cmd- command to handle the request
 */
void server_request(obex_t * handle, obex_object_t * object, int event,
		    int cmd);

/*
 * server_done()
 * Housekeeping event when the server is done servicing.
 * handle- obex handle, object- obex object, obex_cmd- obex command,
 * obex_rsp- obex response
 */
void server_done(obex_t * handle, obex_object_t * object, int obex_cmd,
		 int obex_rsp);

/*
 * fillstream()
 * Method used by the client when empty stream needs to be filled in.
 * handle- obex handle, object- obex object
 * Return the bytes filled
 */
int fillstream(obex_t * handle, obex_object_t * object);

/*
 * connect_server()
 * Routine to connect to the server
 * handle- obex handle, object- obex object
 */
void connect_server(obex_t * handle, obex_object_t * object);

/*
 * get_server()
 * Routine to get objects from the server
 * handle- obex handle, object- obex object
 */
void get_server(obex_t * handle, obex_object_t * object);

/*
 * put_server()
 * This method is responsible for accepting objects
 * handle- obex handle, object- obex object
 */
void put_server(obex_t * handle, obex_object_t * object);

/*
 * easy_readfile()
 * Read a file and alloc a buffer for it
 * filename- name of the file, file_size- size of the file
 * Return the number of bytes read
 */
uint8_t *easy_readfile(const char *filename, int *file_size);

/*
 * get_filesize()
 * Get the filesize in a "portable" way (Taken from obex-apps)
 * filename- name of the file
 * Return size of the file
 */
int get_filesize(const char *filename);

#endif
