/*
 * bt_loc_xml.c: bluetooth location services xml related functions  
 * Copyright 2003 by Columbia University; all rights reserved
 * Author: Howard Chu (hhc42@cs.columbia.edu)
 */

#include "bt_loc_xml.h"

/*
 * xml_process_file()
 * Used by the client to parse the location file received from the server.
 * Right now, it extracts all the elements specified in the location xml
 * file schema and prints them to standard out.
 * file_name- the path of the file to be parsed
 * Returns 0 if succeed, -1 if fails
 */
int xml_process_file(char *file_name)
{
  xmlDocPtr doc;
  xmlNodePtr cur;

  doc = xmlParseFile(file_name);
  if (doc == NULL) {
    fprintf(stderr, "Cannot parse file: %s.\n", file_name);
    return -1;
  }

  cur = xmlDocGetRootElement(doc);
  if (xmlStrcmp(cur->name, (const xmlChar *) LOCATION_DOC_TYPE) != 0) {
    fprintf(stderr, "Invalid location document type\n");
    xmlFreeDoc(doc);
    return -1;
  }

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    xmlChar *tmp;
    if (xmlStrcmp(cur->name, (const xmlChar *) "pos") == 0) {
      /* handle pos */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_pos(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "eme_pos") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_eme_pos(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "trl_pos") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_trl_pos(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "pd") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_pd(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "poserr") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_poserr(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "add_info") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("add_info: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "result") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("result: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "time") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("time: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "alt") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("alt: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "alt_acc") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("alt_acc: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "direction") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("direction: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "speed") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("speed: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "lev_conf") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("lev_conf: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "geo_info") == 0) {
      /* handle geo_info */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_geo_info(doc, cur);
	xmlFree(tmp);
      }
    } else
	if (xmlStrcmp
	    (cur->name,
	     (const xmlChar *) "CoordinateReferenceSystem") == 0) {
      /* handle CoordinateReferenceSystem */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_co_ref_sys(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "Identifier") == 0) {
      /* handle Identifier */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_identifier(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "code") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("code: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "codeSpace") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("codeSpace: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "edition") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("edition: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "categorical") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_categorical(doc, cur);
	xmlFree(tmp);
      }
    }

    cur = cur->next;
  }

  // free the document
  xmlFreeDoc(doc);

  return 0;
}

/*
 * xml_process_buffer
 * Used by the client to parse the location file received from the server.
 * Instead of taking in a file_name and parse the file, it takes in a
 * buffer and stores the buffer in the memory and parse through it.
 * Right now, it extracts all the elements specified in the location xml
 * file schema and prints them to standard out.
 * buffer- the xml buffer to be parsed
 * size- the size of hte buffer
 * Returns 0 if succeed, -1 if fails
 */
int xml_process_buffer(const char *buffer, int size)
{
  xmlDocPtr doc;
  xmlNodePtr cur;

  doc = xmlParseMemory(buffer, size);
  if (doc == NULL) {
    fprintf(stderr, "Cannot parse buffer\n");
    return -1;
  }

  cur = xmlDocGetRootElement(doc);
  if (xmlStrcmp(cur->name, (const xmlChar *) LOCATION_DOC_TYPE) != 0) {
    fprintf(stderr, "Invalid location document type\n");
    xmlFreeDoc(doc);
    return -1;
  }

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    xmlChar *tmp;
    if (xmlStrcmp(cur->name, (const xmlChar *) "pos") == 0) {
      /* handle pos */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_pos(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "eme_pos") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_eme_pos(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "trl_pos") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_trl_pos(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "pd") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_pd(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "poserr") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_poserr(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "add_info") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("add_info: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "result") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("result: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "time") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("time: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "alt") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("alt: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "alt_acc") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("alt_acc: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "direction") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("direction: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "speed") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("speed: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "lev_conf") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("lev_conf: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "geo_info") == 0) {
      /* handle geo_info */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_geo_info(doc, cur);
	xmlFree(tmp);
      }
    } else
	if (xmlStrcmp
	    (cur->name,
	     (const xmlChar *) "CoordinateReferenceSystem") == 0) {
      /* handle CoordinateReferenceSystem */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_co_ref_sys(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "Identifier") == 0) {
      /* handle Identifier */
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	parse_identifier(doc, cur);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "code") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("code: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "codeSpace") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("codeSpace: %s\n", tmp);
	xmlFree(tmp);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "edition") == 0) {
      tmp = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (tmp != NULL) {
	printf("edition: %s\n", tmp);
	xmlFree(tmp);
      }
    }

    cur = cur->next;
  }

  // free the document
  xmlFreeDoc(doc);

  return 0;
}

/*
 * parse_co_ref_sys()
 * parse the coordinate reference system element of the location xml file
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_co_ref_sys(xmlDocPtr doc, xmlNodePtr cur)
{
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;
  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "Identifier") == 0) {
      parse_identifier(doc, cur);
    }
    cur = cur->next;
  }
}

/*
 * parse_geo_info()
 * parse the coordinate reference system element of the location xml file
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_geo_info(xmlDocPtr doc, xmlNodePtr cur)
{
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;
  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "CoordinateReferenceSystem")
	== 0) {
      parse_co_ref_sys(doc, cur);
    }
    cur = cur->next;
  }
}

/*
 * parse_identifier()
 * parse the identifier element
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_identifier(xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "code") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("code: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "codeSpace") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("codeSpace: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "edition") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("edition: %s\n", key);
	xmlFree(key);
      }
    }
    cur = cur->next;
  }
}

/*
 * parse_pos()
 * parse the pos element
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_pos(xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "msid") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("msid: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "pd") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	parse_pd(doc, cur);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "poserr") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	parse_poserr(doc, cur);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "gsm_net_param") ==
	       0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("gsm_net_param: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "x") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("x: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "y") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("y: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "z") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("z: %s\n", key);
	xmlFree(key);
      }
    }

    cur = cur->next;
  }
}

/*
 * parse_eme_pos()
 * parse the eme_pos element
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_eme_pos(xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "msid") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("msid: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "pd") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	parse_pd(doc, cur);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "poserr") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	parse_poserr(doc, cur);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "esrd") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("esrd: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "esrk") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("esrk: %s\n", key);
	xmlFree(key);
      }
    }

    cur = cur->next;
  }
}

/*
 * parse_trl_pos()
 * parse the trl_pos element
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_trl_pos(xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "msid") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("msid: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "pd") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	parse_pd(doc, cur);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "poserr") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	parse_poserr(doc, cur);
	xmlFree(key);
      }
    }
    cur = cur->next;
  }
}

/*
 * parse_pd()
 * parse the pd element
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_pd(xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "time") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("time: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "shape") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("shape: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "alt") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("alt: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "alt_acc") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("alt_acc: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "speed") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("speed: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "direction") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("direction: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "lev_conf") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("lev_conf: %s\n", key);
	xmlFree(key);
      }
    }

    cur = cur->next;
  }
}

/*
 * parse_poserr()
 * parse the poserr element
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_poserr(xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "result") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("result: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "add_info") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("add_info: %s\n", key);
	xmlFree(key);
      }
    } else if (xmlStrcmp(cur->name, (const xmlChar *) "time") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("time: %s\n", key);
	xmlFree(key);
      }
    }
    cur = cur->next;
  }
}

/*
 * parse_categorical()
 * parse the categorical element
 * xmlDocPtr doc- pointer to the xml document
 * xmlNodePtr cur- pointer to the current position in the document
 */
void parse_categorical(xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  printf("%s:\n", cur->name);

  cur = cur->xmlChildrenNode;

  while (cur != NULL) {
    if (xmlStrcmp(cur->name, (const xmlChar *) "type") == 0) {
      key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
      if (key != NULL) {
	printf("type: %s\n", key);
	xmlFree(key);
      }
    }
    cur = cur->next;
  }
}
