/*
 * bt_loc_server.c: bluetooth location services server main program
 * Copyright 2003 by Columbia University; all rights reserved
 * Author: Howard Chu (hhc42@cs.columbia.edu)
 */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/select.h>
#include <sys/stat.h>

#include "bt_loc_lib.h"
#include "bt_loc_obex.h"

int main(int argc, char **argv)
{
  inquiry_info ii[MAX_INQUIRY];
  bdaddr_t bdaddr, dev_bdaddr;
  int file_size;
  char *file_content;
  char tmp[5];
  char dev_bdaddr_str[BDADDR_LEN];
  int dev_id;
  int method;
  obex_t *handle;
  uint8_t channel;
  struct context global_context = { 0, };
  int server_socket;
  FILE *fop;

  // specify what device to be used for easier testing
  if (argc == 3) {
    strcpy(tmp, DEFAULT_DEV_ID);
    dev_id = hci_devid(tmp);

    printf("Using device id: %d\n", dev_id);
    method = atoi(argv[2]);

    if (dev_id == -1 || hci_devba(dev_id, &dev_bdaddr) < 0) {
      perror("Error: Device is not available");
      exit(-1);
    }

    ba2str(&dev_bdaddr, dev_bdaddr_str);
    printf("Device address: %s\n", dev_bdaddr_str);
  } else if (argc == 4) {
    strcpy(tmp, argv[3]);
    dev_id = hci_devid(tmp);

    printf("Using device id: %d\n", dev_id);
    method = atoi(argv[2]);

    if (dev_id == -1 || hci_devba(dev_id, &dev_bdaddr) < 0) {
      perror("Error: Device is not available");
      exit(-1);
    }

    ba2str(&dev_bdaddr, dev_bdaddr_str);
    printf("Device address: %s\n", dev_bdaddr_str);
  } else {
    printf
	("Usage: %s [location xml file] [1 for obex, 2 for sockets] [hci# # is the number of the device (optional)]\n",
	 argv[0]);
    return -1;
  }

  /* check whether the file exists, if it doesn't, exit */
  fop = fopen(argv[1], "r");
  if (fop == NULL) {
    fprintf(stderr, "Error opening location file: %s\n", strerror(errno));
    exit(-1);
  }
  fclose(fop);

  // using l2cap sockets
  if (method == L2CAP_SOCKET) {
    int tmp;

    // search for sdp server, if we can't find it, exit
    if (get_bt_address(dev_id, &bdaddr, ii) < 0) {
      perror("Error: Cannot find server\n");
      exit(-1);
    }
    // if our device doesn't have sdp installed, exit
    if (bacmp(&dev_bdaddr, &bdaddr) != 0) {
      fprintf(stderr, "Error: Device doesn't have SDP installed.\n");
      exit(-1);
    }
    // read the location xml file into a string buffer
    file_size = 0;
    tmp = read_location_file(argv[1], &file_size, &file_content);
    if (tmp == -1) {
      return -1;
    }
    // intialize the server
    if (init_server(&server_socket, dev_bdaddr) < 0) {
      perror("Error: Server initialization failed\n");
      exit(-1);
    }
    // process the request from client
    if (process_request(file_content, file_size, server_socket, argv[1]) <
	0) {
      perror("Error while processing request.\n");
      exit(-1);
    }

    free(file_content);
  }
  // use obex
  else if (method == OBEX) {
    // check whether the file is present first in the current working
    // directory
    int my_channel;
    if (get_opush_address(dev_id, &bdaddr, ii, &my_channel) < 0) {
      perror("Sdp server detection failed.\n");
      exit(-1);
    }
    // if our device doesn't have sdp and opush installed, exit
    if (bacmp(&dev_bdaddr, &bdaddr) != 0) {
      fprintf(stderr, "Device doesn't have SDP installed.\n");
      exit(-1);
    }
    // after the client is disconnected from the server, we need to
    // restart the server to make it start listening again.
    while (1) {
      //channel = BT_CHANNEL;
      channel = my_channel;

      if (!
	  (handle =
	   OBEX_Init(OBEX_TRANS_BLUETOOTH, obex_event_handler, 0))) {
	perror("OBEX_Init failed");
	exit(-1);
      }

      printf("OBEX_Init succeed.\n");

      OBEX_SetUserData(handle, &global_context);

      // register BT server
      if (BtOBEX_ServerRegister(handle, BDADDR_ANY, channel) < 0) {
	perror("Server register error! (Bluetooth)\n");
	exit(-1);
      }

      /* Now process server events */
      printf("OBEX server has started and waits for requests.\n");
      server_do(handle);

      // after the server is done start up the server again.
      continue;
    }
  }
  // invalid method entries
  else {
    printf("Invalid method entries [1 for obex, 2 for sockets]\n");
    exit(-1);
  }

  return 0;
}
